from encrypt import (
    rotate_left,
    rotate_right,
    bytes_to_words,
    words_to_bytes,
)

xor = lambda x, y: x ^ y

def f(block):
    state = bytes_to_words(block)

    for i in range(24):
        old_state = state[:]
        state[0] = rotate_left(old_state[0] ^ old_state[1], 9)
        state[1] = rotate_right(old_state[1] ^ old_state[2], 5)
        state[2] = rotate_right(old_state[2] ^ old_state[3], 3)
        state[3] = old_state[0]

    return words_to_bytes(state)

def inv_f(block):
    state = bytes_to_words(block)

    for i in range(24):
        old_state = state[:]
        state[0] = old_state[3]
        state[1] = rotate_right(old_state[0], 9)  ^ old_state[3]
        state[2] = rotate_left(old_state[1], 5) ^ state[1]
        state[3] = rotate_left(old_state[2], 3) ^ state[2]

    return words_to_bytes(state)

def unpad(b):
    padding = b[-1]
    return b[:-padding]

def decrypt_flag():

    # Content from 0day.py.enc
    zero_day_state = bytes.fromhex("""
        c0c0 1f2a 78c1 f2db be7c 4c09 ca62 00ca
        118c 10bb 7c25 033b 7b10 3ead 3284 9163
        073a bd85 73ac a30a 87f5 55ca 3672 adb5
        5d53 1a35 291f 1c1b dddc 10ae 773f d3df
    """)

    # First 16 bytes from test.py
    init_state      = bytes.fromhex("696d 706f 7274 2072 6571 7565 7374 730a")

    # First 16 bytes from test.py.env
    final_state     = bytes.fromhex("e31a ec64 8133 8430 18f8 abde e58d 3067")

    assert inv_f(f(init_state)) == init_state

    s_inv = inv_f(final_state)

    xor_val = [*map(xor, s_inv, init_state)]

    flag = b""
    for i in range(0, len(zero_day_state), 16):
        block = zero_day_state[i:i+16]
        inv = inv_f(block)
        flag += bytes([*map(xor, xor_val, inv)])
    
    print(unpad(flag))

if __name__ == '__main__':
    decrypt_flag()