#ifndef MMIO_MGT_INCLUDE_H
#define MMIO_MGT_INCLUDE_H
#include <sys/types.h>

#define MMIO_BUFFER_PADDR            ((off_t)0xFE0000000)
#define MMIO_BUFFER_SIZE             ((size_t)0x1000)

#include "list.h"

typedef struct {
    struct list_head list;
    void *vaddr;
    off_t paddr;
    size_t size;
    size_t rounded_size;
} mem_descriptor_t;

/*
 * Open mm_driver.
 *
 * Parameters:
 *      None
 * 
 * Return value:
 *      0 or error code.
 */
int init_mm_driver(void);

/*
 * Fini mm_driver.
 *
 * Parameters:
 *      None
 * 
 * Return value:
 *      0 or error code.
 */
int fini_mm_driver(void);

/*
 * Map physical buffer area.
 *
 * Parameters:
 *      p_ptr       -       pointer to store the virtual address of the buffer
 *      paddr       -       physical address to map
 *      psize       -       physical size to map
 * 
 * Return value:
 *      0 or error code.
 */
int map_physical(void **p_ptr, off_t paddr, size_t psize);

/*
 * Unmap physical buffer area.
 *
 * Parameters:
 *      ptr         -       pointer to unmap
 * 
 * Return value:
 *      0 or error code.
 */
int unmap_physical(void *ptr, size_t size);

/*
 * Map BAR0 buffer area.
 *
 * Parameters:
 *      None
 * 
 * Return value:
 *      0 or error code.
 */
int map_bar0_buffer(void **p_mmio_area);

/*
 * Reserve some physical memory using the mm_driver.
 *
 * Parameter:
 *      size        -       size of the area to allocate
 *      p_paddr     -       pointer to store the physical address
 * 
 * Return value:
 *      0 or error code
 */
int reserve_physical(size_t size, off_t *p_paddr);

/*
 * Release previously allocated physical memory using the mm_driver.
 *
 * Parameter:
 *     paddr        -       physical address of the allocated area
 * 
 * Return value:
 *      0 or error code
 */
int release_physical(off_t paddr);


int allocate_and_map_physical(size_t size, mem_descriptor_t **p_memdesc);

int deallocate_and_unmap_physical(mem_descriptor_t *memdesc);

#endif